import { getNonTypeIdentifiers } from "../util/getNonTypeIdentifiers";
import Transformer from "./Transformer";
/**
 * Class for editing import statements when we are keeping the code as ESM. We still need to remove
 * type-only imports in TypeScript and Flow.
 */
export default class ESMImportTransformer extends Transformer {
    constructor(tokens, isTypeScriptTransformEnabled) {
        super();
        this.tokens = tokens;
        this.isTypeScriptTransformEnabled = isTypeScriptTransformEnabled;
        this.nonTypeIdentifiers = isTypeScriptTransformEnabled
            ? getNonTypeIdentifiers(tokens)
            : new Set();
    }
    process() {
        // TypeScript `import foo = require('foo');` should always just be translated to plain require.
        if (this.tokens.matches3(43024 /* _import */, 2048 /* name */, 14368 /* eq */)) {
            this.tokens.replaceToken("const");
            return true;
        }
        if (this.tokens.matches2(42512 /* _export */, 14368 /* eq */)) {
            this.tokens.replaceToken("module.exports");
            return true;
        }
        if (this.tokens.matches1(43024 /* _import */)) {
            return this.processImport();
        }
        return false;
    }
    processImport() {
        if (this.tokens.matches2(43024 /* _import */, 6144 /* parenL */)) {
            // Dynamic imports don't need to be transformed.
            return false;
        }
        const snapshot = this.tokens.snapshot();
        const allImportsRemoved = this.removeTypeBindings();
        if (allImportsRemoved) {
            this.tokens.restoreToSnapshot(snapshot);
            while (!this.tokens.matches1(1536 /* string */)) {
                this.tokens.removeToken();
            }
            this.tokens.removeToken();
            if (this.tokens.matches1(7680 /* semi */)) {
                this.tokens.removeToken();
            }
        }
        return true;
    }
    /**
     * Remove type bindings from this import, leaving the rest of the import intact.
     *
     * Return true if this import was ONLY types, and thus is eligible for removal. This will bail out
     * of the replacement operation, so we can return early here.
     */
    removeTypeBindings() {
        this.tokens.copyExpectedToken(43024 /* _import */);
        if (this.tokens.matchesContextual(30 /* _type */) &&
            !this.tokens.matchesAtIndex(this.tokens.currentIndex() + 1, [7168 /* comma */]) &&
            !this.tokens.matchesContextualAtIndex(this.tokens.currentIndex() + 1, 10 /* _from */)) {
            // This is an "import type" statement, so exit early.
            return true;
        }
        if (this.tokens.matches1(1536 /* string */)) {
            // This is a bare import, so we should proceed with the import.
            this.tokens.copyToken();
            return false;
        }
        let foundNonTypeImport = false;
        if (this.tokens.matches1(2048 /* name */)) {
            if (this.isTypeName(this.tokens.identifierName())) {
                this.tokens.removeToken();
                if (this.tokens.matches1(7168 /* comma */)) {
                    this.tokens.removeToken();
                }
            }
            else {
                foundNonTypeImport = true;
                this.tokens.copyToken();
                if (this.tokens.matches1(7168 /* comma */)) {
                    this.tokens.copyToken();
                }
            }
        }
        if (this.tokens.matches1(24587 /* star */)) {
            if (this.isTypeName(this.tokens.identifierNameAtIndex(this.tokens.currentIndex() + 2))) {
                this.tokens.removeToken();
                this.tokens.removeToken();
                this.tokens.removeToken();
            }
            else {
                foundNonTypeImport = true;
                this.tokens.copyExpectedToken(24587 /* star */);
                this.tokens.copyExpectedToken(2048 /* name */);
                this.tokens.copyExpectedToken(2048 /* name */);
            }
        }
        else if (this.tokens.matches1(4096 /* braceL */)) {
            this.tokens.copyToken();
            while (!this.tokens.matches1(5120 /* braceR */)) {
                if (this.tokens.matches3(2048 /* name */, 2048 /* name */, 7168 /* comma */) ||
                    this.tokens.matches3(2048 /* name */, 2048 /* name */, 5120 /* braceR */)) {
                    // type foo
                    this.tokens.removeToken();
                    this.tokens.removeToken();
                    if (this.tokens.matches1(7168 /* comma */)) {
                        this.tokens.removeToken();
                    }
                }
                else if (this.tokens.matches5(2048 /* name */, 2048 /* name */, 2048 /* name */, 2048 /* name */, 7168 /* comma */) ||
                    this.tokens.matches5(2048 /* name */, 2048 /* name */, 2048 /* name */, 2048 /* name */, 5120 /* braceR */)) {
                    // type foo as bar
                    this.tokens.removeToken();
                    this.tokens.removeToken();
                    this.tokens.removeToken();
                    this.tokens.removeToken();
                    if (this.tokens.matches1(7168 /* comma */)) {
                        this.tokens.removeToken();
                    }
                }
                else if (this.tokens.matches2(2048 /* name */, 7168 /* comma */) ||
                    this.tokens.matches2(2048 /* name */, 5120 /* braceR */)) {
                    // foo
                    if (this.isTypeName(this.tokens.identifierName())) {
                        this.tokens.removeToken();
                        if (this.tokens.matches1(7168 /* comma */)) {
                            this.tokens.removeToken();
                        }
                    }
                    else {
                        foundNonTypeImport = true;
                        this.tokens.copyToken();
                        if (this.tokens.matches1(7168 /* comma */)) {
                            this.tokens.copyToken();
                        }
                    }
                }
                else if (this.tokens.matches4(2048 /* name */, 2048 /* name */, 2048 /* name */, 7168 /* comma */) ||
                    this.tokens.matches4(2048 /* name */, 2048 /* name */, 2048 /* name */, 5120 /* braceR */)) {
                    // foo as bar
                    if (this.isTypeName(this.tokens.identifierNameAtIndex(this.tokens.currentIndex() + 2))) {
                        this.tokens.removeToken();
                        this.tokens.removeToken();
                        this.tokens.removeToken();
                        if (this.tokens.matches1(7168 /* comma */)) {
                            this.tokens.removeToken();
                        }
                    }
                    else {
                        foundNonTypeImport = true;
                        this.tokens.copyToken();
                        this.tokens.copyToken();
                        this.tokens.copyToken();
                        if (this.tokens.matches1(7168 /* comma */)) {
                            this.tokens.copyToken();
                        }
                    }
                }
                else {
                    throw new Error("Unexpected import form.");
                }
            }
            this.tokens.copyExpectedToken(5120 /* braceR */);
        }
        return !foundNonTypeImport;
    }
    isTypeName(name) {
        return this.isTypeScriptTransformEnabled && !this.nonTypeIdentifiers.has(name);
    }
}
